from ftw.theming.interfaces import ICSSCaching
from ftw.theming.interfaces import ISCSSCompiler
from operator import methodcaller
from plone.app.layout.navigation.root import getNavigationRootObject
from plone.memoize import ram
from plone.memoize.interfaces import ICacheChooser
from Products.CMFCore.utils import getToolByName
from Products.Five import BrowserView
from zope.component import getMultiAdapter
from zope.component import getUtility
from zope.component.hooks import getSite
from zope.interface import alsoProvides
import hashlib


def compute_css_bundle_hash(context):
    """This method creates a hash of the current CSS resource registry.
    We need this in order to detect a bundle recook and flush our caches.
    We do this by using the hash of the currently cooked bundles as part of
    our cache key.
    """
    portal_css = getToolByName(context, 'portal_css')
    theme_resources = dict(
        reduce(list.__add__,
               map(methodcaller('items'),
                   portal_css.concatenatedResourcesByTheme.values())))
    keys = theme_resources.keys()
    map(keys.extend, theme_resources.values())
    return str(hash(frozenset(keys)))


def is_debug_mode_enabled():
    cssregistry = getToolByName(getSite(), 'portal_css')
    return bool(cssregistry.getDebugMode())


def get_css_cache_key(context, debug_mode_caching=True):
    debug_mode_enabled = is_debug_mode_enabled()
    if not debug_mode_caching and debug_mode_enabled:
        return None

    portal = getToolByName(context, 'portal_url').getPortalObject()
    navroot = getNavigationRootObject(context, portal)
    compiler = getMultiAdapter((context, context.REQUEST), ISCSSCompiler)
    key = [navroot.absolute_url(),
           compute_css_bundle_hash(navroot),
           str(navroot.modified().millis()),
           compiler.get_cachekey(dynamic_resources_only=not debug_mode_enabled)]

    return hashlib.md5('.'.join(key)).hexdigest()


def ramcachekey(func, self):
    cachekey = get_css_cache_key(self.context, debug_mode_caching=False)
    if cachekey is None:
        raise ram.DontCache
    return cachekey


def get_theming_css_url(context):
    """Returns the theming.css URL for a context, relative to
    its navigation root.
    Cachekey params may be added.
    """
    portal = getToolByName(context, 'portal_url').getPortalObject()
    navroot = getNavigationRootObject(context, portal)
    css_url = '{0}/theming.css'.format(navroot.absolute_url())
    cachekey = get_css_cache_key(context)
    if cachekey:
        css_url += '?cachekey={0}'.format(cachekey)
    return css_url


class ThemingCSSView(BrowserView):

    def __call__(self):
        response = self.request.response
        response.setHeader('Content-Type', 'text/css; charset=utf-8')
        response.setHeader('X-Theme-Disabled', 'True')
        if self.request.get('cachekey'):
            # Do not set cache headers when no cachekey provided.
            # The cached representation is to be considered fresh for 1 year
            # http://stackoverflow.com/a/3001556/880628
            # The cache header is only active when plone.app.caching is not
            # configured.
            response.setHeader('Cache-Control', 'private, max-age=31536000')
            alsoProvides(self, ICSSCaching)
        return self.get_css()

    @ram.cache(ramcachekey)
    def get_css(self):
        compiler = getMultiAdapter((self.context, self.request), ISCSSCompiler)
        return compiler.compile(debug=is_debug_mode_enabled())

    def invalidate_cache(self):
        cache = getUtility(ICacheChooser)('{0}.get_css'.format(__name__))
        cache.ramcache.invalidateAll()

    def get_url(self):
        return get_theming_css_url(self.context)


class RedirectToNavrootThemingCSSView(BrowserView):
    """The theming.css should only be shipped on navigation roots,
    but there may be requests (e.g. from tinymce) on other contexts.

    By redirecting to the navigation root we can make sure that the
    client-side caching works efficiently.
    """

    def __call__(self):
        return self.request.response.redirect(self.get_url())

    def get_url(self):
        return get_theming_css_url(self.context)



from ftw.builder.testing import BUILDER_LAYER
from ftw.builder.testing import functional_session_factory
from ftw.builder.testing import set_builder_session_factory
from ftw.testing.layer import COMPONENT_REGISTRY_ISOLATION
from ftw.testing.layer import ComponentRegistryLayer
from ftw.testing.layer import TEMP_DIRECTORY
from plone.app.testing import applyProfile
from plone.app.testing import FunctionalTesting
from plone.app.testing import PloneSandboxLayer
from zope.configuration import xmlconfig


class MetaZCMLLayer(ComponentRegistryLayer):

    def setUp(self):
        super(MetaZCMLLayer, self).setUp()
        import ftw.theming
        self.load_zcml_file('meta.zcml', ftw.theming)


META_ZCML = MetaZCMLLayer()


class ThemingLayer(PloneSandboxLayer):
    defaultBases = (COMPONENT_REGISTRY_ISOLATION,
                    BUILDER_LAYER,
                    TEMP_DIRECTORY)

    def setUpZope(self, app, configurationContext):
        xmlconfig.string(
            '<configure xmlns="http://namespaces.zope.org/zope">'
            '  <include package="z3c.autoinclude" file="meta.zcml" />'
            '  <includePlugins package="plone" />'
            '  <includePluginsOverrides package="plone" />'
            '</configure>',
            context=configurationContext)

    def setUpPloneSite(self, portal):
        applyProfile(portal, 'ftw.theming:default')


THEMING_FIXTURE = ThemingLayer()
THEMING_FUNCTIONAL = FunctionalTesting(
    bases=(THEMING_FIXTURE,
           set_builder_session_factory(functional_session_factory)),
    name="ftw.theming:functional")



from ftw.builder import Builder
from ftw.builder import create
from ftw.testbrowser import browsing
from ftw.theming.tests import FunctionalTestCase


class TestIconsAreWrapped(FunctionalTestCase):

    @browsing
    def test_navigation(self, browser):
        self.grant('Manager')
        folder = create(Builder('folder').titled(u'The Folder'))
        create(Builder('file').titled(u'The File').with_dummy_content().within(folder))
        create(Builder('navigation portlet'))

        browser.open(folder)
        navigation = browser.css('.portletNavigationTree')
        file_linke_in_navigation = navigation.find('The File')
        self.assertEquals(
            ['mimetype-icon', 'icon-mimetype-img-doc'],
            file_linke_in_navigation.css('span.mimetype-icon').first.classes)

    @browsing
    def test_folder_contents(self, browser):
        self.grant('Manager')
        folder = create(Builder('folder').titled(u'The Folder'))
        create(Builder('file').titled(u'The File').with_dummy_content().within(folder))

        browser.login().open(folder, view='folder_contents')
        rows = browser.css('#listing-table').first.dicts(as_text=False, head_offset=1)
        title_cell = rows[0][u'Title']
        self.assertEquals(u'The File', title_cell.text)
        self.assertEquals(
            ['mimetype-icon', 'icon-mimetype-img-doc'],
            title_cell.css('span.mimetype-icon').first.classes)



import inspect


def find_object_in_stack(name, klass):
    frame = inspect.currentframe()
    while not isinstance(frame.f_locals.get(name, None), klass):
        frame = frame.f_back
    return frame.f_locals[name]



from ftw.theming.interfaces import IDynamicSCSSResource
from ftw.theming.interfaces import ISCSSFileResource
from ftw.theming.interfaces import ISCSSResource
from ftw.theming.interfaces import SLOTS
from ftw.theming.profileinfo import ProfileInfo
from path import Path
from plone.app.layout.navigation.interfaces import INavigationRoot
from zope.interface import implements
from zope.interface import Interface
import hashlib


class SCSSResource(object):
    """A basic SCSS resource to be extended.
    """
    implements(ISCSSResource)

    def __init__(self, name, slot='addon', before=None, after=None,
                 source=u''):
        if slot not in SLOTS:
            raise ValueError('Invalid slot "{0}". Valid slots: {1}'.format(
                    slot, SLOTS))

        self.name = name
        self.slot = slot
        self.before = before
        self.after = after
        self.source = source

    def available(self, context, request, profileinfo=None):
        return True

    def get_source(self, context, request):
        return self.source

    def get_cachekey(self, context, request):
        raise NotImplementedError(
            'SCSSResource {} has no cachekey'.format(self.name))


class DynamicSCSSResource(SCSSResource):
    """A dynamic SCSS resource provides SCSS source which may change.
    In order to have the correct caching, a cachekey is necessary.

    Dynamic resources should either subclass the DynamicSCSSResource class or
    initialize it with at least a name, a source and a cache key.
    """
    implements(IDynamicSCSSResource)

    def __init__(self, name, slot='addon', before=None, after=None,
                 source=u'', cachekey=None):
        super(DynamicSCSSResource, self).__init__(name=name,
                                                  slot=slot,
                                                  before=before,
                                                  after=after,
                                                  source=source)
        self.cachekey = cachekey

    def get_cachekey(self, context, request):
        if not self.cachekey:
            raise NotImplementedError('No cachekey set.')
        return self.cachekey


class SCSSFileResource(SCSSResource):
    """A scss resource represents a scss file for registering in the scss registry.
    It holds the relevant information for building the scss pipeline.
    """
    implements(ISCSSFileResource)

    def __init__(self, package, relative_path, slot='addon',
                 profile=None, for_=INavigationRoot, layer=Interface,
                 before=None, after=None):
        name = self._make_resource_name(package, relative_path)
        before = before and self._make_resource_name(package, before)
        after = after and self._make_resource_name(package, after)
        super(SCSSFileResource, self).__init__(name, slot=slot,
                                               before=before, after=after)

        self.package = package
        self.relative_path = relative_path
        self.path = self._resolve_path(package, relative_path)
        self.profile = profile
        self.for_ = for_
        self.layer = layer

    def available(self, context, request, profileinfo=None):
        if not self.for_.providedBy(context):
            return False

        if not self.layer.providedBy(request):
            return False

        if self.profile is not None:
            profileinfo = profileinfo or ProfileInfo(context)
            if not profileinfo.is_profile_installed(self.profile):
                return False

        return True

    def get_source(self, context, request):
        return self.path.text('utf-8')

    def get_cachekey(self, context, request):
        return str(self.path.mtime)

    @staticmethod
    def _resolve_path(package, relative_path):
        package_path = Path(__import__(package, fromlist=package).__path__[0])
        absolute_path = package_path.joinpath(relative_path)
        with absolute_path.open():
            pass  # test if file exists
        return absolute_path

    @staticmethod
    def _make_resource_name(package, relative_path):
        if ':' in relative_path:
            return relative_path
        return ':'.join((package, relative_path))



from ftw.theming.interfaces import ISCSSCompiler
from ftw.theming.tests import FunctionalTestCase
from zope.component import getMultiAdapter


class TestThemingExtensions(FunctionalTestCase):

    def test_function_exists(self):
        self.assert_scss_compilation_result(
            'body{rgb-exists:true;foo-exists:false}',

            'body {'
            ' rgb-exists: function-exists(rgb);'
            ' foo-exists: function-exists(foo);'
            '}')

    def assert_scss_compilation_result(self, expected_css, scss_input, msg=None):
        compiler = getMultiAdapter((self.portal, self.request), ISCSSCompiler)
        got_css = compiler.compile_scss_string(scss_input)
        self.assertEquals(expected_css.strip(), got_css.strip(), msg=msg)



from ftw.theming.interfaces import ISCSSRegistry
from ftw.theming.testing import THEMING_FUNCTIONAL
from plone.app.testing import setRoles
from plone.app.testing import TEST_USER_ID
from unittest2 import TestCase
from zope.component import getUtility
import transaction


class FunctionalTestCase(TestCase):
    layer = THEMING_FUNCTIONAL

    def setUp(self):
        self.portal = self.layer['portal']
        self.request = self.layer['request']
        self.registry = getUtility(ISCSSRegistry)
        self.original_resources = self.registry.resources

    def tearDown(self):
        self.original_resources = self.registry.resources

    def grant(self, *roles):
        setRoles(self.portal, TEST_USER_ID, roles)
        transaction.commit()
